# -*- coding: utf-8 -*-

import os
import sys

base_dir = os.path.dirname(__file__)
sys.path.insert(0, base_dir)


from .sup import *
from .net import *
from .args import *
from .work import work_as_sender, work_as_receiver
from .crypto import PycaAES256CBC, PycaFernet
from . import __version__


def main():
    Global.version = __version__

    parser = create_and_init_parser()
    args = parser.parse_args(sys.argv[1:])
    common_parse(args)
    plog("Begin pysharek", 1)

    if args.cipher == 1:
        Global.cipher = PycaFernet()
    elif args.cipher == 2:
        Global.cipher = PycaAES256CBC()
    else:
        pout("Failed successfully (cipher 1 or cipher 2)")
        exit()

    Global.file_dir = args.file[0]
    if args.mode == "send":
        work_as_sender(args)
    elif args.mode == "receive":
        work_as_receiver(args)
    else:
        pout("Failed successfully (send or receive)")
        exit()

if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-

import os
import sys
import io
import json
import datetime
import random


class Global:
    version = None
    outfile = None
    logfile = None
    log_debug = None
    yes_always = None
    file_dir = None
    sock: "socket" = None
    cipher: "PycaAES256CBC or PycaFernet" = None
    file_size_4_message = 262144  # 256 KB  # 1048576  # 1 MB
    hash_4_thread = None
    continue_from = None

    priv_key = None
    pub_key = None
    pub_key_2 = None
    shared_key = None


def pout(msg: str, endl=True):
    if not endl:
        pout_low(msg)
    else:
        pout_low(msg + "\n")


def pout_low(msg: str):
    print(msg, end="")
    if Global.outfile is not None:
        with open(Global.outfile, "a", encoding="utf-8") as fd:
            fd.write(msg)
            fd.flush()
    # if Global.logfile is not None:
    #     with open(Global.logfile, "a", encoding="utf-8") as fd:
    #         fd.write(msg)
    #         fd.flush()
    plog(msg, 5)


def plog(s: str, _type: int = 1):
    prefix = "UNKNOWN"
    if _type == 0:
        prefix = ""
    elif _type == 1:
        prefix = "INFO"
    elif _type == 2:
        prefix = "WARNING"
    elif _type == 3:
        prefix = "ERROR"
    elif _type == 4:
        prefix = "DEBUG"
    elif _type == 5:
        prefix = "OUTPUT"
    elif _type == 6:
        prefix = "SLICING"
    if Global.logfile is not None:
        if not Global.log_debug and prefix == "DEBUG":
            return
        else:
            time_str = datetime.datetime.now().strftime("[%y.%m.%d %H:%M:%S.%f]")
            with open(Global.logfile, "a", encoding="utf-8") as fd:
                fd.write(f"{time_str} ({prefix}) {s}\n")
                fd.flush()


def get_files_list(dir_path: str) -> list:
    return [os.path.join(path, name) for path, subdirs, files in os.walk(dir_path) for name in files]


def get_dirs_needed_for_files(files: list) -> list:
    dirs = set()
    for file_i in files:
        dir_i = os.path.dirname(file_i)
        dirs.add(dir_i)
    dirs = sorted(list(dirs))
    dirs = list(set(dirs))
    return dirs


def mkdir_with_p(path: str):
    os.makedirs(path, exist_ok=True)


def get_file_size(file: str) -> int:
    file = os.path.abspath(file)
    return os.path.getsize(file)
    # return os.stat(file).st_size


def get_dir_size(dir_path: str) -> int:
    files = get_files_list(os.path.abspath(dir_path))
    res = 0
    for file_i in files:
        res += get_file_size(file_i)
    return res


def get_file_time(file: str) -> str:
    # time_str = datetime.datetime.now().strftime("[%y.%m.%d %H:%M:%S.%f]")
    unix_time_stamp = os.path.getmtime(file)
    time_str = datetime.datetime.fromtimestamp(unix_time_stamp).strftime("%y.%m.%d %H:%M:%S.%f")
    return time_str


def write_json(d: dict, path: str):
    s = json.dumps(d)
    with open(path, "w", encoding="utf-8") as fd:
        fd.write(s)


def read_json(path: str) -> dict:
    path = os.path.abspath(path)
    with open(path, "r", encoding="utf-8") as fd:
        s = fd.read()
    return json.loads(s)


def int_to_bytes(a: int) -> bytes:
    if a < 0 or a > 4294967295:
        raise ValueError(f"int_to_bytes: number {a} more than 2**32-1 or less zero. ")
    return a.to_bytes(4, "big")


def bytes_to_int(bs: bytes) -> int:
    res = int.from_bytes(bs, "big")
    if res < 0 or res > 4294967295:
        raise ValueError(f"bytes_to_int: number {res} more than 2**32-1 or less zero from bytes: \"{bs}\". ")
    return res


def utf8_to_bytes(s: str) -> bytes:
    return s.encode("utf-8")


def bytes_to_utf8(bs: bytes) -> str:
    return str(bs, "utf-8")


def get_random_string(_lenght : int = 20) -> str:
    import string
    S = ''.join(random.choices(string.ascii_uppercase + string.digits, k=_lenght))
    return S


def get_nice_size(size_bytes: int) -> str:
    if size_bytes < 1024:
        return f"{size_bytes} B"
    elif size_bytes < 1024*1024:
        devider = 1024
        return f"{round(size_bytes / devider, 3)} KB"
    elif size_bytes < 1024*1024*1024:
        devider = 1024*1024
        return f"{round(size_bytes / devider, 3)} MB"
    elif size_bytes < 1024*1024*1024*1024:
        devider = 1024*1024*1024
        return f"{round(size_bytes / devider, 3)} GB"
    else:
        devider = 1024*1024*1024*1024
        return f"{round(size_bytes / devider, 3)} TB"



# -*- coding: utf-8 -*-

import hashlib
import os

from alive_progress import alive_bar

from .sup import get_files_list, get_file_size, get_dir_size


def calc_hash_bytes(bs: bytes) -> str:
    return hashlib.sha256(bs).hexdigest()


def calc_hash_str(s: str) -> str:
    return hashlib.sha256(s.encode("utf-8")).hexdigest()


def calc_hash_file(path: str) -> str:
    path = os.path.abspath(path)

    block_size = 65536  # 64 kB
    sha = hashlib.sha256()
    with open(path, "rb") as fd:
        file_buffer = fd.read(block_size)
        while len(file_buffer) > 0:
            sha.update(file_buffer)
            file_buffer = fd.read(block_size)
    return sha.hexdigest()


# https://github.com/rsalmei/alive-progress/issues/20
def calc_hash_dir(dir_path: str, if_hierarchy=True) -> str:
    dir_path = os.path.abspath(dir_path)
    files = get_files_list(dir_path)
    files = sorted(files)
    files_rel = [os.path.relpath(file_i, dir_path) for file_i in files]
    sha = hashlib.sha256()
    if if_hierarchy:
        for file_i in files_rel:
            sha.update(file_i.encode("utf-8"))

    dir_size = get_dir_size(dir_path)
    hashes = []
    with alive_bar(dir_size) as bar:
        for file_i in files:
            file_i_size = get_file_size(file_i)
            file_i_hash = calc_hash_file(file_i)
            hashes.append(file_i_hash)
            bar(file_i_size)

    hashes = sorted(hashes)

    for hash_i in hashes:
        sha.update(hash_i.encode("utf-8"))

    return sha.hexdigest()



# -*- coding: utf-8 -*-

import argparse
from .sup import Global


def create_and_init_parser() -> "argparse.ArgumentParser":
    parser = argparse.ArgumentParser(prog="pysharek",
                                     description="Help to secure send file between two peers.")
    parser.add_argument("file", type=str, nargs=1,
                        help="Path to source/output file/directory")

    parser.add_argument("--dublicate_out_to_file", type=str, default=None, required=False,
                        help="Duplicate program output to file")

    parser.add_argument("--log_file", type=str, default=None, required=False,
                        help="Duplicate program output to file")

    parser.add_argument("--log_debug", default=False, required=False, action='store_true',
                        help="if \"log_file\" defined, lvl DEBUG will be logged")

    parser.add_argument("--version", action="version", version=f"V{Global.version}",
                        help="Check version of pysharek")

    parser.add_argument("--connect", type=str, choices=["server", "client"], required=True,
                        help="Usage as server or client")

    parser.add_argument("--mode", type=str, choices=["send", "receive"], required=True,
                        help="Usage as sender or receiver")

    parser.add_argument("--cipher", type=int, choices=[1, 2], required=False, default=1,
                        help="Choose what cipher will be used. PycaFernet (1) or PycaAES256CBC (2). Default 1")

    parser.add_argument("--ip", type=str, default=None,
                        help="If connect=\"client\", define ip to connect")

    parser.add_argument("--port", type=int, required=True,
                        help="Define port for bind/connect")

    parser.add_argument("--password", type=str, required=False,
                        help="Define password for encryption while transferring. "
                             "Password must be same for the sender and receiver")
    parser.add_argument("--yes", default=False, required=False, action="store_true",
                        help="Answer \"yes\" on any questions")
    parser.add_argument("--continue_from", type=str, required=False, default=None,
                        help="continue from {file}:{block}.")

    return parser


def common_parse(args: "argparse.Namespace") -> None:
    Global.outfile = args.dublicate_out_to_file
    Global.logfile = args.log_file
    Global.log_debug = args.log_debug
    Global.yes_always = args.yes

    if args.continue_from is not None:
        continue_f, continue_s = args.continue_from.strip().lower().split(":")
        continue_f, continue_s = int(continue_f), int(continue_s)
        Global.continue_from = (continue_f, continue_s)
    else:
        Global.continue_from = None



# -*- coding: utf-8 -*-

import sys
import os
import io
import json
import argparse
import hashlib
import threading
import time
import traceback

from .sup import *
from .net import *
from .hashes_work import calc_hash_file
from .hashes_work import calc_hash_dir
from .sup import get_dirs_needed_for_files
from .crypto import transfer_shared_key


def work_as_sender(args: "argparse.Namespace"):
    # python -m pysharek --mode send --connect server --port 1337
    #                           --password 123 /tmp/123.mp4 --log_file /tmp/log2.txt --log_debug
    plog("I am sender", 1)
    common_block_of_sender_and_receiver(args)
    file = Global.file_dir

    if os.path.isfile(file):
        send_file(file)
    elif os.path.isdir(file):
        send_dir(file)
    elif not os.path.exists(file):
        pout(f"No such file \"{file}\". Exiting")
        Global.sock.close()
        exit()
    else:
        pout(f"Failed successfully (what is file \"{file}\")")
        Global.sock.close()
        exit()


def calc_hash_4_thread(path: str, dir_file: bool):
    if dir_file == False:
        calculated_hash = calc_hash_dir(path, True)
    else:
        calculated_hash = calc_hash_file(path)
    Global.hash_4_thread = calculated_hash


def send_file(file_name: str):
    file = os.path.abspath(file_name)
    plog(f"File (not dir) \"{file}\" will be sended", 1)
    sock = Global.sock
    meta = build_file_meta(file)
    plog(f"meta={meta}", 4)
    plog("Sending meta...", 1)
    send_crypto_msg(sock, meta, b"")
    plog("Meta sended!", 1)
    with open(file, "rb") as fd:
        file_size = meta["file"]["size"]
        plog(f"File size {file_size} bytes", 4)
        readed = 0
        with alive_bar(file_size) as bar:
            block_size = Global.file_size_4_message
            file_buffer = fd.read(block_size)
            plog(f"Readed {len(file_buffer)} bytes", 4)
            bar(len(file_buffer)-1)  # I do not know why need -1?!
            readed += len(file_buffer)
            while len(file_buffer) > 0:
                while True:
                    js_send = {"type": "sending", "file_count": 1, "slice": readed}
                    plog(f"Sendeding: {js_send} + data", 4)
                    send_crypto_msg(sock, js_send, file_buffer)
                    plog(f"Sended", 4)
                    js, bs = recv_crypto_msg(sock)
                    plog(f"Received js={js}", 4)
                    if "type" not in js or js["type"] != "received":
                        pout("Cannot send block. Trying again.")
                    else:
                        break
                file_buffer = fd.read(block_size)
                bar(len(file_buffer))
                readed += len(file_buffer)

    pout(f"Calculating hash...")
    file_hash = calc_hash_file(file)
    pout(f"\"{file_hash}\" is hash of file=\"{file}\"")
    plog(f"Sending file hash", 1)
    hash_msg = {"type": "hash", "file_num": 1, "hash": f"{file_hash}"}
    plog(f"Hash message={hash_msg}", 4)
    send_crypto_msg(sock, hash_msg, b"")
    plog(f"Hash message sended", 4)


def send_dir(dir_name: str):
    file_gi, slice_gi = 0, 0
    try:
        dir_name = os.path.abspath(dir_name)
        plog(f"Dir \"{dir_name}\" will be sended", 1)
        meta = build_dir_meta(dir_name)
        files = sorted(meta["files"].keys())
        sock = Global.sock
        plog(f"meta={meta}", 4)
        plog("Sending meta...", 1)
        send_crypto_msg(sock, meta, b"")
        plog("Meta sended!", 1)
        dir_size = meta["dir_size"]
        pout(f"{get_nice_size(dir_size)} will be sended")
        with alive_bar(dir_size) as bar:
            for file_i in files:
                file_gi += 1
                slice_gi = 0
                file_size = meta["files"][file_i]["size"]
                if Global.continue_from is not None:
                    if file_gi < Global.continue_from[0]:
                        bar(file_size)
                        continue
                file = os.path.join(dir_name, file_i)
                with open(file, "rb") as fd:
                    if Global.continue_from is not None:
                        if file_gi < Global.continue_from[0]:
                            skip_bytes = Global.file_size_4_message*Global.continue_from[1]
                            fd.seek(skip_bytes)
                            bar(skip_bytes)
                    plog(f"File size {file_size} bytes", 4)
                    readed = 0
                    block_size = Global.file_size_4_message
                    file_buffer = fd.read(block_size)
                    slice_gi += 1
                    bar(len(file_buffer)-1)  # I do not know why need -1?!
                    readed += len(file_buffer)
                    while len(file_buffer) > 0:
                        while True:
                            js_send = {"type": "sending", "file_num": file_gi, "slice": readed}
                            plog(f"Sending: {js_send} + data", 4)
                            send_crypto_msg(sock, js_send, file_buffer)
                            plog(f"Sended", 4)
                            js, bs = recv_crypto_msg(sock)
                            plog(f"Received js={js}", 4)
                            if "type" not in js or js["type"] != "received":
                                pout("Cannot send block. Trying again.")
                            else:
                                break
                        file_buffer = fd.read(block_size)
                        slice_gi += 1
                        plog(f"{file_gi}:{slice_gi}", 6)
                        bar(len(file_buffer))
                        readed += len(file_buffer)
    except Exception as e:  # TODO: catch CTRL+C
        pout(f"Error: {e}")
        traceback.print_exc()
        pout(f"Last slising: {file_gi}:{slice_gi}")
        exit(1)

    pout(f"Calculating hash...")
    dir_hash = calc_hash_dir(dir_name, True)
    pout(f"\"{dir_hash}\" is hash of dir \"{dir_name}\"")
    plog(f"Sending file hash", 1)
    hash_msg = {"type": "hash", "hash": f"{dir_hash}"}
    plog(f"Hash message={hash_msg}", 4)
    send_crypto_msg(sock, hash_msg, b"")
    plog(f"Hash message sended", 4)


def work_as_receiver(args: "argparse.Namespace"):
    # python -m pysharek --mode receive --connect client --ip 127.0.0.1 --port 1337
    #                               --password 123 /tmp/333/1234.mp4 --log_file /tmp/log1.txt --log_debug
    plog("I am receiver", 1)
    common_block_of_sender_and_receiver(args)
    file = Global.file_dir

    plog("Receiving meta", 1)
    meta, trash = recv_crypto_msg(Global.sock)
    plog("Received", 1)
    plog(f"meta={meta}", 4)
    if "mode" not in meta:
        pout("Error while receive meta")
        Global.sock.close()
        exit()
    if meta["mode"] == 1:
        receive_file(file, meta)
    elif meta["mode"] == 2:
        receive_dir(file, meta)
    else:
        pout(f"Failed successfully (what is mode \"{meta['mode']}\"?)")


def receive_file(file_name: str, meta: dict):
    plog("File (not dir) will be received", 1)
    sock = Global.sock
    file = os.path.abspath(file_name)
    if os.path.isfile(file):
        pout(f"File \"{file}\" already exists. Delete it? (Y/n) ", endl=False)
        user_in = input()
        plog(f"User enter={user_in}", 4)
        if not (user_in == "" or user_in[0].lower() in ["y", "1"]):
            pout(f"Not delete. Exiting...")
            Global.sock.close()
            exit()
    if os.path.isdir(file):
        pout(f"Tt is planned to receive a file, and this \"{file}\" is a directory. Exiting...")
        Global.sock.close()
        exit()

    file_dir = os.path.dirname(file)
    if not os.path.exists(file_dir):
        plog(f"Making dir: {file_dir}", 4)
        mkdir_with_p(file_dir)

    with open(file_name, "wb") as fd:
        writed = 0
        file_size = meta["file"]["size"]
        plog(f"Receiving file size={file_size}", 1)
        with alive_bar(file_size) as bar:
            while writed != file_size:
                while True:
                    plog("Receiving file block", 4)
                    js, file_buffer = recv_crypto_msg(sock)
                    plog(f"Received file block {len(file_buffer)} bytes, js={js}", 4)
                    if "type" not in js or js["type"] != "sending":
                        pout(f"Cannot receive. Requesting the block again.")
                        send_crypto_msg(sock, {"type": "error_again"}, b"")
                        continue
                    writed += len(file_buffer)
                    if js["slice"] != writed:
                        pout(f"Slice {js['slice']} dont same with writed {writed}. Exiting")
                        Global.sock.close()
                        exit()
                    answer = {"type": "received"}
                    plog(f"Sending answer={answer}", 4)
                    send_crypto_msg(sock, answer, b"")
                    plog(f"Sended", 4)
                    fd.write(file_buffer)
                    fd.flush()
                    bar(len(file_buffer))
                    break
    pout(f"Receiving and calculating hash...")

    x = threading.Thread(target=calc_hash_4_thread, args=(file, True))
    x.start()

    js, trash = recv_crypto_msg(sock)
    if "type" not in js or js["type"] != "hash":
        pout(f"Cannot receive hash. Exiting")
        Global.sock.close()
        exit()
    recv_hash = js["hash"]
    pout(f"\"{recv_hash}\" is received hash")
    # file_hash = calc_hash_file(file)
    while Global.hash_4_thread is None:
        time.sleep(1.0)  # Dude, what is join?
    file_hash = Global.hash_4_thread
    pout(f"\"{file_hash}\" is hash of file=\"{file}\"")
    pout(f"Comparing hashes: \"{recv_hash}\" (recieved) vs \"{file_hash}\" (calculating)")
    if recv_hash != file_hash:
        pout(f"{'='*15} HASHES DOES NOT MATCH!!! {'='*15}")
    else:
        pout(f"Hashes matched. All is OK!")


def receive_dir(dir_name: str, meta: dict):
    dir_name = os.path.abspath(dir_name)
    if Global.yes_always == False:
        if len(get_files_list(dir_name)) != 0:
            user_in = input(f"Directory \"{dir_name}\" is not empty!!! Continue? (Y): ")
            if user_in.strip().lower() not in ["", "y", "1", "yes", "yep"]:
                exit()

    file_gi, slice_gi = 0, 0
    try:
        plog(f"Dir will be received", 1)
        sock = Global.sock
        files = sorted(meta["files"].keys())
        files_4_dirs = [os.path.join(dir_name, file_i) for file_i in files]
        needed_dirs = get_dirs_needed_for_files(files_4_dirs)
        for needed_dir_i in needed_dirs:
            mkdir_with_p(needed_dir_i)
        dir_size = meta["dir_size"]
        pout(f"{get_nice_size(dir_size)} will be sended")
        with alive_bar(dir_size) as bar:
            for file_i in files:
                file_gi += 1
                slice_gi = 0
                file_size = meta["files"][file_i]["size"]
                if Global.continue_from is not None:
                    if file_gi < Global.continue_from[0]:
                        bar(file_size)
                        continue
                file_name = os.path.join(dir_name, file_i)
                with open(file_name, "wb") as fd:
                    if Global.continue_from is not None:
                        if file_gi < Global.continue_from[0]:
                            skip_bytes = Global.file_size_4_message * Global.continue_from[1]
                            fd.seek(skip_bytes)
                            bar(skip_bytes)
                    writed = 0
                    plog(f"Receiving file size={file_size}", 1)
                    while writed != file_size:
                        while True:
                            plog("Receiving file block", 4)
                            js, file_buffer = recv_crypto_msg(sock)
                            plog(f"Received file block {len(file_buffer)} bytes, js={js}", 4)
                            if "type" not in js or js["type"] != "sending":
                                pout(f"Cannot receive. Requesting the block again.")
                                send_crypto_msg(sock, {"type": "error_again"}, b"")
                                continue
                            writed += len(file_buffer)
                            if js["slice"] != writed:
                                pout(f"Slice {js['slice']} dont same with writed {writed}. Exiting")
                                Global.sock.close()
                                exit()
                            answer = {"type": "received"}
                            plog(f"Sending answer={answer}", 4)
                            send_crypto_msg(sock, answer, b"")
                            plog(f"Sended", 4)
                            fd.write(file_buffer)
                            plog(f"{file_gi}:{slice_gi}", 6)
                            slice_gi += 1
                            fd.flush()
                            bar(len(file_buffer))
                            break
    except Exception as e:  # TODO: catch CTRL+C
        pout(f"Error: {e}")
        traceback.print_exc()
        pout(f"Last slising: {file_gi}:{slice_gi}")
        exit(1)


    pout(f"Receiving and calculating hash...")

    x = threading.Thread(target=calc_hash_4_thread, args=(dir_name, False))
    x.start()

    js, trash = recv_crypto_msg(sock)
    if "type" not in js or js["type"] != "hash":
        pout(f"Cannot receive hash. Exiting")
        Global.sock.close()
        exit()
    recv_hash = js["hash"]
    pout(f"\n\"{recv_hash}\" is received hash")
    # dir_hash = calc_hash_dir(dir_name, True)
    while Global.hash_4_thread is None:
        time.sleep(1.0)  # Dude, what is join?
    dir_hash = Global.hash_4_thread
    pout(f"\"{dir_hash}\" is hash of dir=\"{dir_name}\"")
    pout(f"Comparing hashes: \"{recv_hash}\" (recieved) vs \"{dir_hash}\" (calculating)")
    if recv_hash != dir_hash:
        pout(f"{'='*15} HASHES DOES NOT MATCH!!! {'='*15}")
    else:
        pout(f"Hashes matched. All is OK!")


def common_block_of_sender_and_receiver(args: "argparse.Namespace"):
    plog(f"Common part achieved", 1)
    sock = init_socks_by_connect_and_do_handshake(args)
    plog(f"Handshaked", 1)
    Global.sock = sock

    if args.password is None:
        from cryptography.hazmat.primitives.asymmetric.x448 import X448PublicKey
        from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey
        from cryptography.hazmat.primitives import serialization

        Global.priv_key = X448PrivateKey.generate()
        Global.pub_key = Global.priv_key.public_key()
        Global.pub_key = Global.pub_key.public_bytes(encoding=serialization.Encoding.Raw,
                                                     format=serialization.PublicFormat.Raw)
        pub_key_2 = None
        if args.connect == "server":  # TODO: remove args.* == *
            send_msg(sock, {"key_exchange": "1"}, Global.pub_key)
            d, pub_key_2 = recv_msg(sock)
            if "key_exchange" not in d or d["key_exchange"] != "2":
                pout("Cannot key_exchange-2")
                plog("Cannot key_exchange-2")
                socket_close(sock)
                exit()
        elif args.connect == "client":
            d, pub_key_2 = recv_msg(sock)
            if "key_exchange" not in d or d["key_exchange"] != "1":
                pout("Cannot key_exchange-1")
                plog("Cannot key_exchange-1")
                socket_close(sock)
                exit()
            else:
                send_msg(sock, {"key_exchange": "2"}, Global.pub_key)
        Global.pub_key_2 = pub_key_2
        Global.shared_key = Global.priv_key.exchange(X448PublicKey.from_public_bytes(Global.pub_key_2))
        Global.shared_key = transfer_shared_key(Global.shared_key)
        Global.cipher.set_password(Global.shared_key)
    else:
        Global.cipher.set_password(args.password)
    Global.cipher.start()

    plog(f"Inited cipher in common part", 1)

    challenge_cipher(sock)


def init_socks_by_connect_and_do_handshake(args: "argparse.Namespace") -> socket:
    plog(f"It was as {args.connect}", 1)
    plog(f"Init socker and making handshake", 1)
    if args.connect == "server":
        sock = socket_create_and_listen(args.port)
        handshake_as_server(sock)
    elif args.connect == "client":
        if args.ip is None:
            pout("If connect=\"client\", \"--ip\" must be defined")
            exit()
        sock = socket_create_and_connect(args.ip, args.port)
        handshake_as_client(sock)
    else:
        pout("Failed successfully (server or client)")
        exit()
    return sock


def challenge_cipher(sock: socket):
    plog("Start cipher challenge", 1)
    num1 = bytes_to_int(os.urandom(4))
    num2 = bytes_to_int(os.urandom(4))
    num3 = num1+num2
    d = {"num1": num1, "num2": num2, "num3": num3}
    plog(f"Sending formed challenge={d}", 4)
    send_crypto_msg(sock, d, b"")
    plog(f"Sended", 1)

    d, trash = recv_crypto_msg(sock)
    plog(f"Received challenge={d}", 4)
    if "num1" not in d or "num2" not in d or "num3" not in d or d["num1"] + d["num2"] != d["num3"]:
        pout("(challenge_cipher): Cannot challenge")
        socket_close(sock)
        exit()
    plog("Cipher challenged", 1)


def build_dir_meta(dir_path: str) -> dict:
    dir_path = os.path.abspath(dir_path)
    files = get_files_list(dir_path)
    files = [os.path.relpath(file_i, dir_path) for file_i in files]
    files = sorted(files)

    d = {"mode": 2, "files": {}}
    for file_i in files:
        src_file_i = os.path.join(dir_path, file_i)
        d["files"][file_i] = {"size": get_file_size(src_file_i), "time": get_file_time(src_file_i)}

    dir_size = 0
    for file_i in d["files"]:
        dir_size += d["files"][file_i]["size"]
    d["dir_size"] = dir_size

    return d


def build_file_meta(file_name: str) -> dict:
    file_name = os.path.abspath(file_name)
    d = dict({"mode": 1})
    d["file"] = {"size": get_file_size(file_name), "time": get_file_time(file_name)}
    return d



# -*- coding: utf-8 -*-

# pip3 install build
# python3 -m build
# python3 -m pip install --upgrade twine
# python3 -m twine upload --repository testpypi dist/*
# pip3 install pysharek --no-deps --index-url https://test.pypi.org/simple

import os
import sys

base_dir = os.path.dirname(__file__)
sys.path.insert(0, base_dir)


from .main import main


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import json
import hashlib
import os
import time
from alive_progress import alive_bar

from .sup import bytes_to_int, int_to_bytes, plog, Global, pout
import socket
from .crypto import PycaAES256CBC


def print_bytes(bs: bytes):
    res = ""
    for i in bs:
        res += f"{i}_"
    res = res[:-1]
    print(f"\n{res}")


def send_msg(conn, js: dict, bs: bytes):
    msg_hash_len = 32  # sha256
    js = json.dumps(js).encode("utf-8")
    js_len, bs_len = len(js), len(bs)
    js_len_b, bs_len_b = int_to_bytes(js_len), int_to_bytes(bs_len)
    msg = js_len_b + js + bs_len_b + bs
    msg_hash = hashlib.sha256(msg).digest()
    msg_len_b = int_to_bytes(len(msg) + msg_hash_len)
    msg = msg_len_b + msg + msg_hash

    buff = conn.send(msg)


def send_crypto_msg(conn, js: dict, bs: bytes):
    assert Global.cipher.is_started()

    msg_hash_len = 32  # sha256
    js = json.dumps(js).encode("utf-8")
    js_len, bs_len = len(js), len(bs)
    js_len_b, bs_len_b = int_to_bytes(js_len), int_to_bytes(bs_len)
    msg = js_len_b + js + bs_len_b + bs

    msg = Global.cipher.encrypt(msg)
    msg_hash = hashlib.sha256(msg).digest()

    msg_len_b = int_to_bytes(len(msg) + msg_hash_len)
    msg = msg_len_b + msg + msg_hash

    buff = conn.send(msg)


def recv_msg(conn) -> (dict, bytes):
    msg_hash_len = 32  # sha256
    msg_size = conn.recv(4, socket.MSG_WAITALL)
    msg_size = bytes_to_int(msg_size)
    msg = conn.recv(msg_size, socket.MSG_WAITALL)
    js_size = bytes_to_int(msg[:4])
    js = msg[4:4+js_size]
    js = json.loads(js.decode("utf-8"))
    bs_size = bytes_to_int(msg[4+js_size:4+js_size+4])
    bs = msg[4+js_size+4:4+js_size+4+bs_size]
    msg_hash = msg[4+js_size+4+bs_size:]
    control_hash = hashlib.sha256(msg[:4+js_size+4+bs_size]).digest()
    if msg_hash != control_hash:
        return None
    else:
        return (js, bs)


def recv_crypto_msg(conn) -> (dict, bytes):
    assert Global.cipher.is_started()

    msg_hash_len = 32  # sha256
    msg_size = conn.recv(4, socket.MSG_WAITALL)
    msg_size = bytes_to_int(msg_size)
    msg_and_hash = conn.recv(msg_size, socket.MSG_WAITALL)
    msg, msg_hash = msg_and_hash[:-msg_hash_len], msg_and_hash[-msg_hash_len:]
    control_hash = hashlib.sha256(msg).digest()

    msg = Global.cipher.decrypt(msg)

    js_size = bytes_to_int(msg[:4])
    js = msg[4:4+js_size]
    js = json.loads(js.decode("utf-8"))
    bs_size = bytes_to_int(msg[4+js_size:4+js_size+4])
    bs = msg[4+js_size+4:4+js_size+4+bs_size]
    # control_hash = hashlib.sha256(msg[:4+js_size+4+bs_size]).digest()
    if msg_hash != control_hash:
        return None
    else:
        return (js, bs)


def socket_create_and_connect(ip: str, port: int) -> socket:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while True:
        try:
            sock.connect((ip, port))
        except ConnectionRefusedError:
            wait_time = 3
            plog(f"(socket_create_and_connect) ConnectionRefusedError, waiting {wait_time}")
            time.sleep(wait_time)
            continue
        break
    return sock


def socket_create_and_listen(port: int) -> socket:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", port))
    sock.listen(1)
    conn, info = sock.accept()
    return conn


def socket_close(sock: socket):
    sock.close()


def pand(bs: bytes, n: int) -> bytes:
    _n = len(bs)
    if _n % n != 0:
        res = bs + b'\x00'*(n - _n % n)
    else:
        res = bs
    return res


def handshake_as_server(sock: socket):
    d, iv = recv_msg(sock)
    if "handshake" not in d or d["handshake"] != "1":
        plog("(handshake_as_server): Cannot handshake 1", 3)
        socket_close(sock)
        exit()
    check = hashlib.sha256("handshake".encode("utf-8") + iv).digest()
    send_msg(sock, {"handshake": "2"}, check)
    d, check2 = recv_msg(sock)
    if "handshake" not in d or d["handshake"] != "3":
        plog("(handshake_as_server): Cannot handshake 2", 3)
        socket_close(sock)
        exit()
    if check2 != b"OK":
        plog("(handshake_as_server): not b\"OK\" msg received", 3)
        socket_close(sock)
        exit()
    iv = hashlib.sha256(iv).digest()[:16]
    Global.cipher.set_iv(iv)


def handshake_as_client(sock: socket):
    iv = os.urandom(16)
    send_msg(sock, {"handshake": "1"}, iv)
    check = hashlib.sha256("handshake".encode("utf-8") + iv).digest()
    d, check2 = recv_msg(sock)
    if "handshake" not in d or d["handshake"] != "2":
        plog("(handshake_as_client): Cannot handshake 1", 3)
        socket_close(sock)
        exit()
    if check != check2:
        plog("(handshake_as_client): hashes does not match!", 3)
        socket_close(sock)
        exit()
    else:
        send_msg(sock, {"handshake": "3"}, b"OK")
    iv = hashlib.sha256(iv).digest()[:16]
    Global.cipher.set_iv(iv)


def test_net_1():
    import sys
    if len(sys.argv) == 2:
        file = sys.argv[1]
        sock = socket_create_and_connect("127.0.0.1", 8881)
        with open(file, "rb") as fd:
            bs = fd.read()
        send_msg(sock, {"msg": "hello"}, bs)
    else:
        sock = socket_create_and_listen(8881)
        js, bs = recv_msg(sock)
        print(js)
        with open("/tmp/test_file.bin", "wb") as fd:
            fd.write(bs)
            fd.flush()

    socket_close(sock)


def test_net_2():
    import sys
    if sys.argv[1] == "1":
        sock = socket_create_and_connect("127.0.0.1", 8882)
        handshake_as_client(sock)
    else:
        sock = socket_create_and_listen(8882)
        handshake_as_server(sock)

    socket_close(sock)



# -*- coding: utf-8 -*-

import base64
import hashlib
import os

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes, padding
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from sup import get_random_string


class PycaAES256CBC:
    """
    # First, init:
    ``` python
    cipher = PycaAES256CBC()
    ```

    # Second set password and iv:
    ``` python
    cipher.set_passwort("Your strong🤡 utf-8 password")
    cipher.set_iv(os.urandom(16))
    ```

    # Third start cipher:
    ``` python
    cipher.start()
    ```

    # Fourth use:

    ``` python
    text = "Secret message"
    encrypted_message = cipher.encrypt(text.encode("utf-8"))
    decrypted_message = cipher.decrypt(encrypted_message)
    assert decrypted_message.decode("utf-8") == text
    ```
    """

    def __init__(self):
        self.iv = None
        self.key = None
        self.cipher = None

    def set_password(self, password: str):
        pwd = password.encode("utf-8")
        salt = hashlib.sha256(pwd).digest()[:16]
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=480000
        )
        # self.key = base64.urlsafe_b64encode(kdf.derive(pwd))
        self.key = kdf.derive(pwd)

    def set_iv(self, iv: bytes):
        self.iv = iv

    def start(self):
        mod_iv = hashlib.sha256(self.iv + hashlib.sha256(self.key).digest()).digest()[:16]
        self.cipher = Cipher(algorithms.AES(self.key), modes.CBC(mod_iv))

    def is_started(self) -> bool:
        return self.cipher is not None

    def encrypt(self, bs: bytes) -> bytes or None:
        if self.is_started():
            bs = inject_salt(bs)
            padder = padding.PKCS7(algorithms.AES256.block_size).padder()
            bs = padder.update(bs) + padder.finalize()

            encryptor = self.cipher.encryptor()
            ct = encryptor.update(bs) + encryptor.finalize()
            return ct
        else:
            return None

    def decrypt(self, ct: bytes) -> bytes or None:
        if self.is_started():
            decryptor = self.cipher.decryptor()
            bs = decryptor.update(ct) + decryptor.finalize()

            unpadder = padding.PKCS7(algorithms.AES256.block_size).unpadder()
            bs = unpadder.update(bs) + unpadder.finalize()
            bs = takeout_salt(bs)
            return bs
        else:
            return None


# https://github.com/pyca/cryptography/issues/3446
# https://cryptography.io/en/latest/hazmat/primitives/symmetric-encryption/#cryptography.hazmat.primitives.ciphers.Cipher
# https://cryptography.io/en/latest/fernet/#using-passwords-with-fernet


class PycaFernet:
    """
    # First, init:
    ``` python
    cipher = PycaFernet()
    ```

    # Second set password and iv:
    ``` python
    cipher.set_passwort("Your strong🤡 utf-8 password")
    cipher.set_iv(os.urandom(16))  # =)
    ```

    # Third start cipher:
    ``` python
    cipher.start()
    ```

    # Fourth use:

    ``` python
    text = "Secret message"
    encrypted_message = cipher.encrypt(text.encode("utf-8"))
    decrypted_message = cipher.decrypt(encrypted_message)
    assert decrypted_message.decode("utf-8") == text
    ```
    """

    def __init__(self):
        self.iv = None
        self.key = None
        self.cipher = None

    def set_password(self, password: str):
        # key_len = len(Fernet.generate_key())
        pwd = password.encode("utf-8")
        salt = hashlib.sha256(pwd).digest()[:16]
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,  # key_len
            salt=salt,
            iterations=480000
        )
        # self.key = base64.urlsafe_b64encode(kdf.derive(pwd))
        self.key = kdf.derive(pwd)
        self.key = base64.urlsafe_b64encode(self.key)

    def set_iv(self, iv):
        pass  # =)

    def start(self):
        self.cipher = Fernet(self.key)

    def is_started(self) -> bool:
        return self.cipher is not None

    def encrypt(self, bs: bytes) -> bytes or None:
        if self.is_started():
            return self.cipher.encrypt(bs)
        else:
            return None

    def decrypt(self, ct: bytes) -> bytes or None:
        if self.is_started():
            return self.cipher.decrypt(ct)
        else:
            return None


# https://stackoverflow.com/questions/66476150/fast-concatenation-of-bytes-in-python3
def inject_salt(data: bytes) -> bytes:
    salt_prop = 4
    if len(data) == 0:
        # print(0, 1, 1)
        return os.urandom(1)
    salt_slots = 1 + len(data) // salt_prop
    salts = os.urandom(salt_slots)
    len_res = len(data) + salt_slots
    # print(len(data), salt_slots, len_res)
    res = bytearray(len_res)
    i, j, k = 0, 0, 0
    while j < len(res):
        if i % salt_prop == 0:
            res[j] = salts[k]
            j += 1
            k += 1
            if j >= len(res):
                break
        res[j] = data[i]
        i += 1
        j += 1
    return bytes(res)


def takeout_salt(data: bytes) -> bytes:
    salt_prop = 4
    # data_slots = ((len(data)-1)*salt_prop)//(salt_prop+1)
    data_slots = (len(data)-1) % (salt_prop+1)
    data_slots = ((len(data)-1-data_slots)//(salt_prop+1))*salt_prop + data_slots
    # salt_slots = 1 + data_slots // salt_prop
    # print(data_slots, salt_slots, len(data))
    res = bytearray(data_slots)
    i, j, c = 0, 1, 1
    while i < len(res):
        res[i] = data[j]
        if c >= salt_prop:
            j += 1
            c = 0
        c += 1
        i += 1
        j += 1
    return bytes(res)


def transfer_shared_key(shared_key: bytes) -> str:
    return hashlib.sha256(shared_key).hexdigest()


def __test_salt():
    import random
    from alive_progress import alive_bar
    # v, k = 10**9, 256
    # v, k = 10**9, 10**3
    v, k = 10**9, 250000
    # v, k = 10**9, 10**6
    n = v//k
    with alive_bar(n) as bar:
        for i in range(n):
            data_i = os.urandom(random.randint(k-100, k+100))
            salted_i = inject_salt(data_i)
            unsalted_i = takeout_salt(salted_i)
            assert data_i == unsalted_i
            bar()


def __test_cipher():
    import random
    from alive_progress import alive_bar
    cipher = PycaAES256CBC()
    # cipher = PycaFernet()
    cipher.set_iv(os.urandom(16)), cipher.set_password(get_random_string())
    cipher.start()
    v, k = 10 ** 9, 250000
    n = v//k
    with alive_bar(n) as bar:
        for i in range(n):
            data_i = os.urandom(random.randint(k - 100, k + 100))
            data_en = cipher.encrypt(data_i)
            data_de = cipher.decrypt(data_en)
            assert data_i == data_de
            bar()


def __test_key_exchange():
    from cryptography.hazmat.primitives.asymmetric.x448 import X448PublicKey
    from cryptography.hazmat.primitives.asymmetric.x448 import X448PrivateKey
    from cryptography.hazmat.primitives import serialization

    private_key1 = X448PrivateKey.generate()
    public_key1 = private_key1.public_key()
    public_key1_bytes = public_key1.public_bytes(encoding=serialization.Encoding.Raw,
                                                 format=serialization.PublicFormat.Raw)
    print(f"Pub key 1 bytes: {public_key1_bytes}")
    public_key1_text = base64.b64encode(public_key1_bytes).decode("ascii")
    print(f"Pub key 1: {public_key1_text}")

    private_key2 = X448PrivateKey.generate()
    public_key2 = private_key2.public_key()
    public_key2_bytes = public_key2.public_bytes(encoding=serialization.Encoding.Raw,
                                                 format=serialization.PublicFormat.Raw)

    print(f"Pub key 2 bytes: {public_key2_bytes}")
    public_key2_text = base64.b64encode(public_key2_bytes).decode("ascii")
    print(f"Pub key 2: {public_key2_text}")

    # =============================

    pubkey1_bytes = base64.b64decode(public_key1_text.encode("ascii"))
    pubkey1 = X448PublicKey.from_public_bytes(pubkey1_bytes)

    pubkey2_bytes = base64.b64decode(public_key2_text.encode("ascii"))
    pubkey2 = X448PublicKey.from_public_bytes(pubkey2_bytes)

    shared_key1 = private_key2.exchange(pubkey1)
    shared_key1 = base64.b64encode(shared_key1).decode("ascii")
    print(f"shared_key1: {shared_key1}")

    shared_key2 = private_key1.exchange(pubkey2)
    shared_key2 = base64.b64encode(shared_key2).decode("ascii")
    print(f"shared_key2: {shared_key2}")


if __name__ == "__main__":
    # __test_salt()
    # __test_cipher()
    __test_key_exchange()
